﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading.Tasks;
using ClientServer.Db;
using ClientServer.Helpers;
using ClientServer.Models;
using Microsoft.AspNetCore.Http;
using Microsoft.AspNetCore.Mvc;
using Microsoft.EntityFrameworkCore;

namespace ClientServer.Controllers.Core.Exercises
{
    /// <summary>
    /// a controller for the plangs
    /// </summary>
    [Route(Constants.ApiPrefix + "plangs")]
    public class PLangController : ControllerWithDb
    {
        public PLangController(YapexDbContext context) : base(context)
        {
        }

        private async Task<bool> _CheckPermission()
        {
            if (!await base.IsLoggedIn()) return false;

            //user needs permission

            if (base.HasSystemPermission(role => role != null && role.SystemRolePermission.CanChangeSystemSettings) ==
                false)
            {
                await
                    Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.NoPermission, "no permission")));
                return false;
            }

            return true;
        }

        /// <summary>
        /// returns all known plangs
        /// (internal name is required for the forntend editor to support syntax highlighting)
        /// </summary>
        [HttpGet()]
        public async Task GetAll()
        {
            if (!await base.IsLoggedIn()) return;

            var pLangs = await _context.PLangs
                .OrderBy(p => p.DisplayName)
                .Select(p => new PLangFullBase()
                {
                    Id = p.Id,
                    DisplayName = p.DisplayName,
                    InternalName = p.InternalName,
                    EditorHighlightModeName = p.EditorHighlightModeName,
                    FileExtensionsWithDot = p.FileExtensionsWithDot
                })
                .ToListAsync();

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponseWithData<List<PLangFullBase>>(ResponseCode.Ok,
                    "plangs queried", pLangs)));
        }

        /// <summary>
        /// returns all known plangs with some extra data
        /// user needs permission to change plang to get this
        /// extra data is: can plang be deleted?
        /// </summary>
        [HttpGet("get/extra")]
        public async Task GetWithExtraData()
        {
            if (!await base.IsLoggedIn()) return;

            List<PLang> pLangs = await _context.PLangs
                .OrderBy(p => p.DisplayName)
                .ToListAsync();


            var forFrontend = pLangs.Select((p) => new PLangWithExtrasFromBackend()
            {
                Id = p.Id,
                InternalName = p.InternalName,
                DisplayName = p.DisplayName,
                EditorHighlightModeName = p.EditorHighlightModeName,
                FileExtensionsWithDot = p.FileExtensionsWithDot,
                CanBeDeleted = false,
            }).ToList();

            foreach (var pLangForFrontend in forFrontend)
            {
                var reasonCannotBeDeleted = await this._GetCanPLangBeDeleted(pLangForFrontend.Id);
                pLangForFrontend.CanBeDeleted = reasonCannotBeDeleted == null;
                pLangForFrontend.CannotBeDeletedReason = reasonCannotBeDeleted;
            }


            await
                Response.WriteAsync(Jc.Serialize(new BasicResponseWithData<List<PLangWithExtrasFromBackend>>(
                    ResponseCode.Ok,
                    "", forFrontend)));
        }

        /// <summary>
        /// creates a new plang
        /// </summary>
        /// <param name="pLang"></param>
        /// <returns>the created plang</returns>
        [HttpPost("create")]
        public async Task CreatePLang([FromBody] PLangFullBase pLang)
        {
            if (await this._CheckPermission() == false) return;

            if (await this.ValidatePLang(pLang) == false) return;


            var newPLang = new PLang()
            {
                DisplayName = pLang.DisplayName,
                InternalName = pLang.InternalName,
                EditorHighlightModeName = pLang.EditorHighlightModeName,
                FileExtensionsWithDot = pLang.FileExtensionsWithDot
            };

            try
            {
                _context.PLangs.Add(newPLang);
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }

            pLang.Id = newPLang.Id;


            await
                Response.WriteAsync(Jc.Serialize(new BasicResponseWithData<PLangFullBase>(ResponseCode.Ok,
                    "", pLang)));
        }

        /// <summary>
        /// updates the given plang 
        /// </summary>
        /// <param name="pLang"></param>
        /// <returns>the updated plang</returns>
        [HttpPost("update")]
        public async Task UpdatePLang([FromBody] PLangFullBase pLang)
        {
            if (await this._CheckPermission() == false) return;


            PLang oldPLang = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == pLang.Id);

            if (oldPLang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no old plang found")));
                return;
            }

            if (await this.ValidatePLang(pLang) == false) return;

            oldPLang.DisplayName = pLang.DisplayName;
            oldPLang.InternalName = pLang.InternalName;
            oldPLang.EditorHighlightModeName = pLang.EditorHighlightModeName;
            oldPLang.FileExtensionsWithDot = pLang.FileExtensionsWithDot;

            try
            {
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }

            var reasonCannotBeDeleted = await _GetCanPLangBeDeleted(oldPLang.Id);

            var forFrontend = new PLangWithExtrasFromBackend()
            {
                Id = oldPLang.Id,
                DisplayName = oldPLang.DisplayName,
                InternalName = oldPLang.InternalName,
                EditorHighlightModeName = oldPLang.EditorHighlightModeName,
                FileExtensionsWithDot = oldPLang.FileExtensionsWithDot,
                CanBeDeleted = reasonCannotBeDeleted == null,
                CannotBeDeletedReason = reasonCannotBeDeleted,
            };

            await
                Response.WriteAsync(Jc.Serialize(new BasicResponseWithData<PLangWithExtrasFromBackend>(ResponseCode.Ok,
                    "", forFrontend)));
        }


        /// <summary>
        /// deletes the given plang by id
        /// </summary>
        /// <param name="pLangId"></param>
        /// <returns></returns>
        [HttpDelete("delete/{pLangId}")]
        public async Task DeletePLang(int pLangId)
        {
            if (await this._CheckPermission() == false) return;


            PLang oldPLang = await _context.PLangs
                .FirstOrDefaultAsync(p => p.Id == pLangId);

            if (oldPLang == null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "no old plang found")));
                return;
            }

            //check if we can delete plang... only if we have no release with this plang & no code template

            var reason = await this._GetCanPLangBeDeleted(oldPLang.Id);

            if (reason != null)
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.InvalidRequest, reason)));
                return;
            }

            try
            {
                _context.PLangs.Remove(oldPLang);
                await _context.SaveChangesAsync();
            }
            catch (Exception e)
            {
                await base.HandleDbError(e);
                return;
            }


            await Response.WriteAsync(Jc.Serialize(new BasicResponse(ResponseCode.Ok, "")));
        }


        private async Task<bool> ValidatePLang(PLangFullBase pLang)
        {
            if (string.IsNullOrEmpty(pLang.DisplayName))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "display name was empty")));
                return false;
            }

            if (string.IsNullOrEmpty(pLang.InternalName))
            {
                await
                    Response.WriteAsync(
                        Jc.Serialize(new BasicResponse(ResponseCode.NotFound, "internal name was empty")));
                return false;
            }

            return true;
        }

        /// <summary>
        /// 
        /// </summary>
        /// <param name="pLangId"></param>
        /// <returns>null: all ok, other the reason why the plang cannot be deleted</returns>
        protected async Task<string> _GetCanPLangBeDeleted(int pLangId)
        {
            var releaseCount = await _context.ExerciseReleases
                    .CountAsync(p => p.PLangId == pLangId)
                ;

            if (releaseCount > 0)
            {
                return "there are still releases with this plang";
            }


            var exerciseInGroupCount = await _context.Exercises
                    .Where(p => p.IsOnlyVisibleToOwner == false &&
                                p.CodeTemplates.Any(k => k.PLangId == pLangId))
                    .ToListAsync()
                ;

            if (exerciseInGroupCount.Count > 0)
            {
                return
                    $"there are {exerciseInGroupCount.Count} exercises with code templates that use the given plang." +
                    $" the first exercise is '{exerciseInGroupCount.First().DisplayName}'";
            }

            var exerciseFromUser = await _context.Exercises
                    .Include(p => p.User)
                    .Where(p => p.IsOnlyVisibleToOwner &&
                                p.CodeTemplates.Any(k => k.PLangId == pLangId))
                    .ToListAsync()
                ;


            if (exerciseFromUser.Count > 0)
            {
                var firstExercise = exerciseFromUser.First();
                return
                    $"there are {exerciseFromUser.Count} exercises (that are only visible to the owner) with code templates that use the given plang." +
                    $" the first exercise is '{firstExercise.DisplayName}' from the user '{firstExercise.User.LastName}, {firstExercise.User.FirstName} (token: {firstExercise.User.Token})'";
            }

            return null;
        }
    }

    public class PLangFullBase
    {
        public int Id { get; set; }

        /// <summary>
        /// the name of the programming language
        /// </summary>
        public string DisplayName { get; set; }

        /// <summary>
        /// the internal name e.g. used for the online-editor
        /// ALSO make sure it's a valid file path would be good to use no spaces... c++ -> cpp
        /// </summary>
        public string InternalName { get; set; }

        /// <summary>
        /// a comma separated list of file extensions for this plang
        /// </summary>
        public string FileExtensionsWithDot { get; set; }

        /// <summary>
        /// the highlight mode for the editor (currently ace)
        /// </summary>
        public string EditorHighlightModeName { get; set; }
    }

    public class PLangWithExtrasFromBackend : PLangFullBase
    {
        /// <summary>
        /// true: plang can be deleted without any doubts (without deleting other data)
        ///   this is true if we have 0 releases with this plang AND there are no code templates for this plang
        /// false: deleting plang would cascade and delete other data 
        /// </summary>
        public bool CanBeDeleted { get; set; }

        /// <summary>
        /// the reason why the plang cannot be deleted
        /// can be null if CanBeDeleted is true
        /// </summary>
        public string CannotBeDeletedReason { get; set; }
    }
}
